;;; Avoid a conflict with the package :GL of SBCL's CLX that was introduced
;;; with the GLX extensions support of SBCL CLX version 0.6.1.
(delete-package 'gl)
(require :cube-modeller)

(in-package :cl-user)

(defun get-pathnames-from-file (&optional (source-file #p"/etc/ld.so.conf"))
  (flet ((string-to-pathname (string)
	   (unless (string-equal "/" (subseq string (1- (length string))))
	     (setf string (concatenate 'string string "/")))
	   (parse-namestring string)))
  (let (paths)
    (with-open-file (file source-file)
      (do ((line (read-line file) (read-line file nil 'eof)))
	  ((eq line 'eof))
	(unless (string= line "")
	  (pushnew (string-to-pathname line) paths))))
    paths)))

(defun get-library-paths ()
  (append '(#p"/lib/" #p"/usr/lib/" #p"/usr/local/lib/")
	  '(#p"/usr/lib/cl-sdl/")
	  (get-pathnames-from-file)
	  (list *default-pathname-defaults*)))

(defparameter *library-paths* nil)

(defparameter *library-types* '("so" "a" "o" "0")) ;; "0" is HACKHACK

(defun find-foreign-library (names &key (library-paths *library-paths* library-paths-p)
			                (library-types *library-types*))
  (unless *library-paths*
    (setf *library-paths* (get-library-paths))
    (unless library-paths-p
      (setf library-paths *library-paths*)))
  (let (pathname)
    (dolist (name names)
      (dolist (path library-paths)
	(dolist (type library-types)
	  (setf pathname (make-pathname :name name
					:type type
					:directory (etypecase path
						     (pathname (pathname-directory path))
						     (list path)
						     (string (pathname-directory
							      (parse-namestring path))))))
	  (when (probe-file pathname)
	    (return-from find-foreign-library pathname))))))
  nil)


(in-package "SB-ALIEN")

(let ((function (symbol-function 'try-reopen-shared-object)))
  (setf (symbol-function 'try-reopen-shared-object)
	#'(lambda (obj)
	    (declare (type shared-object obj))

	    (format  *trace-output* ">>> 1) old shared-object file: ~d~%" (shared-object-file obj))
	    (setf (shared-object-file obj)
		  (namestring (cl-user::find-foreign-library
			       (list
				(if (and (string= (pathname-name (shared-object-file obj))
						  "alien")
					 (string= (car (last (pathname-directory (shared-object-file obj))))
						  "sb-bsd-sockets"))
				    "alien-bsd"
				    (pathname-name (shared-object-file obj)))))))
	    (format  *trace-output* ">>> 2)    now taking the file: ~d now.~%" (shared-object-file obj))
	    
	    (funcall function obj))))

(in-package :cl-user)
(sb-ext:save-lisp-and-die "cube-modeller.core"
			  :toplevel (lambda ()
                                      (format t "~&~%=============== LED Cube Modeller ===============~&~%")
                                      
				      (setf cube::*device-file*
                                            (sb-ext:posix-getenv "CUBETTY")

                                            cube-modeller::*animation-pathname*
                                            (merge-pathnames (make-pathname :directory '(:relative "data"))
                                                             (cl-fad:pathname-as-directory (SB-UNIX:POSIX-GETCWD))))
				      (when cube::*device-file*
					(format t "~&Writing cube data to device file \"~d\".~&~%" cube::*device-file*))
				      (cube-modeller:start)
				      (sb-ext:quit)))
