(in-package :cube-modeller)

;;; (defparameter *perspective-specification* '(45d0 1d0 0.5d0 10d0))
(defparameter *perspective-specification* '(35d0 1d0 0.5d0 10d0))

(defparameter *spin* 0d0)
(defparameter *move-x* 0d0)
(defparameter *move-z* 0d0)
(defparameter *x-old* 0f0)
(defparameter *y-old* 0f0)
(defparameter *alpha-default* -8f0)
(defparameter *beta-default* -36f0)
(defparameter *alpha* *alpha-default*)
(defparameter *beta* *beta-default*)
(defparameter *rotate-mode* nil)

(defparameter *width* nil)
(defparameter *height* nil)

;;; LEDs
(defparameter *LED-diameter* .04d0)
(defparameter *LED-distance* .20f0)

(defvar *modeller-window-shown* nil)

(defun init-gl ()
  (gl:clear-color 1f0 1f0 1f0 0f0)
  (gl:matrix-mode gl:+projection+)
  (gl:load-identity)
  ;; (gl:ortho -1d0 1d0 -1d0 1d0 0d0 10d0)
  (apply #'glu:perspective *perspective-specification*)
  
  (gl:enable gl:+depth-test+)
  (gl:enable gl:+lighting+)
  (gl:enable gl:+light0+))  

(defun draw-sphere (size)
  (glu:with-quadric (sphere)
    (glu:quadric-draw-style sphere glu:+fill+)
    (glu:quadric-normals sphere glu:+smooth+)
    (glu:sphere sphere size 25 4))) ;; 20)))

(defun display-scene ()
  (gl:clear (logior gl:+color-buffer-bit+
		    gl:+depth-buffer-bit+))

  (gl:matrix-mode gl:+modelview+)
  (gl:load-identity)

  ;; define light source
  (sgum:with-single-float-values (ambient 0.3f0 0.3f0 0.3f0 1f0)
    (gl:light-fv gl:+light0+ gl:+ambient+ ambient))
  (sgum:with-single-float-values (diffuse 0.9f0 0.9f0 0.9f0 0f0)
    (gl:light-fv gl:+light0+ gl:+diffuse+ diffuse))
  (sgum:with-single-float-values (specular 1f0 1f0 1f0 1f0)
    (gl:light-fv gl:+light0+ gl:+specular+ specular))
  (sgum:with-single-float-values (position 1f0 1f0 1f0 1f0)
    (gl:light-fv gl:+light0+ gl:+position+ position))

  (glu:look-at 0d0 2d0 1.58d0 0d0 0d0 0d0 0d0 1d0 0d0)
  ;; (glu:look-at 0.5d0 1d0 1.58d0 0d0 0d0 0d0 0d0 1d0 0d0)

  (gl:rotate-f *beta* 1f0 0f0 0f0)
  (gl:rotate-f *alpha* 0f0 1f0 0f0)

  ;; define LED material
  (sgum:with-single-float-values (emission 0f0 0f0 0f0 0f0)
    (gl:material-fv gl:+front+ gl:+emission+ emission))
  (sgum:with-single-float-values (ambient 0.2f0 0.2f0 0.2f0 1f0)
    (gl:material-fv gl:+front+ gl:+ambient+ ambient))
  (sgum:with-single-float-values (specular 0.15f0 0.15f0 0.15f0 1f0)
    (gl:material-fv gl:+front+ gl:+specular+ specular))
  (sgum:with-single-float-values (shininess 15f0)
    (gl:material-fv gl:+front+ gl:+shininess+ shininess))
  
  (gl:normal-3f 0f0 1f0 0f0)

  ;; LEDs

  (gl:init-names)
  (gl:push-name 0)

      (gl:push-matrix)
        (sgum:with-single-float-values (diffuse 0f0 0.5f0 0f0 1f0)
	  (gl:material-fv gl:+front+ gl:+diffuse+ diffuse))
	(gl:translate-f 0f0 (* 3 *LED-distance*) 0f0)
	(draw-sphere (/ *LED-DIAMETER* 2))
      (gl:pop-matrix)

    (let ((origin (* -2 *LED-distance*)))
      (gl:translate-f (- origin) origin (- origin)))
    (dotimes (x 5)
      (dotimes (z 5)
	(dotimes (y 5)
	  (let* ((LEDs-intensity (aref (current-cube) x y z))
		 (grey-value  (* (- 5 (- (char-code LEDs-intensity) 48))
				 0.2f0)))
	    (if (eq LEDs-intensity #\0)
		;; LED is off
		(sgum:with-single-float-values (diffuse 1f0 1f0 1f0 1f0)
		  (gl:material-fv gl:+front+ gl:+diffuse+ diffuse))
		;; LED is on
		(sgum:with-single-float-values (diffuse 0.9f0 grey-value grey-value 1f0)
		  (gl:material-fv gl:+front+ gl:+diffuse+ diffuse))
		))
	  (gl:load-name (+ (* (+ (* 5 z) y) 5) x)) ;; LED id
  	    (draw-sphere *LED-DIAMETER*)
	  (gl:end) ;; of gl:load-name
	  (gl:translate-f (- *LED-distance*) 0f0 0f0))
	(gl:translate-f (* 5 *LED-distance*) *LED-distance* 0f0))
      (gl:translate-f 0f0 (* -5 *LED-distance*) (- *LED-distance*)))

  (sdl:gl-swap-buffers))

(defun get-LED-number (x y)
  "The algorithm of this function was taken from the Linux/SDL port of the tutorial
  \"Object Selection\" by Ben Humphrey (digiben@gametutorials.com, date: 7/26/01); the port
  was done by gOLeM (golem@gametutorials.com) and insane (evilsmile@gametutorials.com).
  see http://www.gametutorials.com/Tutorials/opengl/OpenGL_Pg3.htm

  This algorithm is plainly wrong, but there is no problem unless you attach more than 
  one name to any object. I've to correct this at some time.."
  (let* ((number-of-selected-objects 0)
	   ;; the number of hit records copied to the select-buffer array
	 (number-of-objects-to-store 6)
           ;; the number of hit records that should fit into the select-buffer array
	   ;; It should equal the number of objects that might be selected with one click.
	   ;; As we have to compare the depth of all these objects to get the closest one, this
	   ;; has right now to be six (five LED spheres and the green sphere). Veeerry nice.. :/
	 viewport-coordinates
	   ;; array to store the viewport rectangle specified by:
	   ;; 1st: x- and 2nd: y-coordinate of the lower-left corner,  3rd: width, and 4th: height
	 select-buffer
	   ;; store up to number-of-objects-to-store "hit records"; for each one there are 4 elements:
	   ;;   1st: "the number of names in the name stack at the time of the event"
	   ;;   2nd: "the minimum depth value of all vertices that hit since the previous event"
	   ;;   3rd: "the maximum depth value of all vertices that hit since the previous event"
	   ;;   4th: "the name stack contents, bottom name first."
	   ;; "Depth values (which are in the range [0,1]) are multiplied by 2^32 - 1,
           ;;  before being placed in the hit record."
           ;; source: glselectbuffer(3) - Linux man page
	   ;; "Returned depth values are mapped such that the largest unsigned integer value corresponds
	   ;;  to window coordinate depth 1.0, and zero corresponds to window coordinate depth 0.0."
	   ;; source: The Microsoft Developer Network (MSDN)
	 )

    (labels ((make-viewport-coordinates-array ()
	       (sgum:allocate-foreign-object 'gl:int 4))
	     (access-element-of-the-viewport-coordinates-array (element-index)
	       (sgum:deref-array viewport-coordinates '(:array gl:int) element-index))
	       
	     (make-select-buffer-array (number-of-objects-to-store)
	       (sgum:allocate-foreign-object 'gl:uint (* number-of-objects-to-store 4)))
	     (access-element-of-the-select-buffer-array (element-index)
	       (sgum:deref-array select-buffer '(:array gl:uint) element-index))
	     (get-depth-of-select-buffer-object (object-number)
	       (access-element-of-the-select-buffer-array (1+ (* object-number 4))))
	     (get-id-of-select-buffer-object (object-number)
	       (access-element-of-the-select-buffer-array (+ (* object-number 4) 3))))

      (setf viewport-coordinates (make-viewport-coordinates-array)
	    select-buffer (make-select-buffer-array number-of-objects-to-store))
      
      (gl:select-buffer (* number-of-objects-to-store 4) select-buffer)
      
      (gl:get-integerv gl:+viewport+ viewport-coordinates)
      
      (gl:matrix-mode gl:+projection+)
      (gl:push-matrix)
        (gl:render-mode gl:+select+)
        (gl:load-identity)
        (glu:pick-matrix (coerce x 'double-float)
			 (coerce (- (access-element-of-the-viewport-coordinates-array 3) y)
				 'double-float)
			 2d0 2d0 ;; look in a square of 2x2 pixel
			 viewport-coordinates)
	(apply #'glu:perspective *perspective-specification*)
	(gl:matrix-mode gl:+modelview+)
	(display-scene)
	(setf number-of-selected-objects (gl:render-mode gl:+render+))
	(gl:matrix-mode gl:+projection+)
      (gl:pop-matrix)
      (gl:matrix-mode gl:+modelview+)

;;;      (format t "~d objects found~%" number-of-selected-objects) (force-output)

     ;; number-of-selected-objects contains "number of hit records copied to" the select-buffer array.
     ;; If the number of hits was larger than there is space for hit records in the select-buffer array,
     ;; "a negative hit record count" is stored in number-of-selected-objects.  As we have to compare
     ;; the depth of all these objects to get the closest one, the select-buffer is useless if it was
     ;; overflown. Therefore we can only determine the actually wanted object if a positive number has
     ;; been stored in number-of-selected-objects.
      (when (> number-of-selected-objects 0)
	(let ((lowest-depth (expt 2 32)) ;; most positive uint + 1 (see above)
	      number-of-closest-object)
	  (dotimes (i number-of-selected-objects)
	    (let ((current-depth (get-depth-of-select-buffer-object i)))
	      (when (< current-depth lowest-depth)
		(setf lowest-depth current-depth
		      number-of-closest-object i))))
;;;	  (format t "id of closest object: ~d~%" (get-id-of-select-buffer-object number-of-closest-object)) (force-output)
	  (get-id-of-select-buffer-object number-of-closest-object))))))

(defun mouse-motion (x y)
  (when *rotate-mode*
    (incf *beta* (* 0.2f0
		    (- y *y-old*)))
    (incf *alpha* (* 0.2f0
		     (- x *x-old*)))
    (setf *x-old* x
	  *y-old* y)))
  
(defun mouse (button state x y)
  (setf *x-old* x
	*y-old* y
	*rotate-mode* (and (eq state :down)
			   (eq button :right)))
  (when (and (eq state :down)
	     (eq button :left))
    (let ((led-number (get-LED-number x y)))
      (when led-number
	(destructuring-bind (x y z)
	    (led-number-to-coord led-number)
;;;	  (format t "-> ~d, ~d ~d ~d <-~%" led-number x y z) (force-output)
	  (set-LED x y z)
	  (write-cube-to-device))))))

(defun handle-keypress (key)
  (if (member key ;; Why stop at all? (mgr, 20060102)
              (list 273 274 275 276 (char-code #\s) (char-code #\x)
                    (char-code #\a) (char-code #\b)
                    (char-code #\e) (char-code #\f) (char-code #\d) (char-code #\k)))
        (stop-animation))

  (cond
    ;; move the states of the LEDs within the current cube
    ((eql 273 key) ; up arrow
     (move-forward))
    ((eql 274 key) ; down arrow
     (move-backward))
    ((eql 276 key) ; left arrow
     (move-left))
    ((eql 275 key) ; right arrow
     (move-right))
    ((eql (char-code #\s) key) ; s
     (move-up))
    ((eql (char-code #\x) key) ; x
     (move-down))
    
    ;; change *current-intensity*
    ((and (<= key (char-code #\5))
          (>= key (char-code #\0)))
     (set-current-intensity (code-char key)))
    
    ;; switch to previous/next cube
    ((eql 280 key) ;; page-up
     (goto-cube :next))
    ((eql 281 key) ;; page-down
     (goto-cube :previous))
    ((eql 278 key) ;; home
     (goto-cube :first))
    ((eql 279 key) ;; end
     (goto-cube :last))
    
    ;; insert a copy of the current-cube after or before it
    ((eql (char-code #\a) key)
     (insert-cube-copy))
    ((eql (char-code #\b) key)
     (insert-cube-copy :position :before))
    
    ;; fill the current-cube
    ((eql (char-code #\f) key)
     (fill-cube))

    ;; empty the current-cube
    ((eql (char-code #\e) key)
     (empty-cube))
    
    ;; delete the current-cube
    ((eql (char-code #\d) key)
     (delete-cube))
          
    ;; delete (kill) the whole animation
    ((eql (char-code #\k) key)
     (delete-animation))
          
    ;; reset perspective
    ((eql (char-code #\r) key)
     (reset-perspective))

    ;; start/stop animation (toggle status)
    ((eql (char-code #\p) key)
     (toggle-animation-running-state))

    ((eql (char-code #\u) key)
     (undo))

;;; old animation code	    
;;;      (let ((old-current-cube-number *current-cube-number*))
;;; ;	(do () (nil)
;;; ;       (dotimes (n 100)
;;; 	  (dotimes (i (length *animation*))
;;; 	    (setf *current-cube-number* i)
;;; 	    (display-scene)
;;; 	    (write-cube-to-device)
;;; 	    (sleep-for-delay)))
;;; 	(setf *current-cube-number* old-current-cube-number)
;;; 	(write-cube-to-device)))

    (t
     (format t "~&Unbound key code: ~a (~a)~%" key (sdl:get-key-name key)))))


(defun run-sdl-event-loop ()
  (sdl:event-loop
   (:key-down (key)
	      (if (= key (char-code #\q))
		  (return)
		  (handle-keypress key)))
   (:mouse-button-up (button x y)
;;;                     (format t "Mouse button up: ~A (~A, ~A)~%" button x y)
		     (mouse button :up x y))
   (:mouse-button-down (button x y)
;;;                       (format t "Mouse button dn: ~A (~A, ~A)~%" button x y)
		       (mouse button :down x y))
   (:mouse-motion  (x y xrel yrel state)
		   (declare (ignore xrel yrel state))
		   (mouse-motion x y))
   (:quit ()
	  (return))
   (:resize (width height)
	    (format t "Resized width = ~A height = ~A~%" width height)
	    (init-sdl-with-gl :width width :height height
                              :show-info nil))
   (:idle ()
          (animation-step)
	  (display-scene))))

(defun start-viewer ()
  (if *modeller-window-shown*
      (format t "start-viewer: The cube modeller OpenGL window is already been displayed.~%              Just returning without doing anything more.~%")
      (unwind-protect
	   (progn
	     (setf *modeller-window-shown* t)
	     
	     (init-sdl-with-gl :show-info nil)

	     ;; ??  (gl:depth-func gl:+lequal+)
	     (init-gl)
             ;; (goto-cube :first)
	     ;; (write-cube-to-device)
	     (run-sdl-event-loop))
	(progn
	  (sdl:quit)
          (close-cube-device)
	  (setf *modeller-window-shown* nil)))))


(defun init-sdl-with-gl (&key (width 700) (height 700) (bpp 16) (show-info t)
			 (video-flags (logior sdl:+opengl+ sdl:+resizable+ sdl:+gl-doublebuffer+));; sdl:+hwpalette+))
			 )
  (setf *width* width
	*height* height)
  (sdl:init sdl:+init-video+)
  (sdl:set-gl-attributes :red-size 4 :blue-size 4 :green-size 4
			 :doublebuffer 1 :depth-size 16)
  (sdl:gl-set-attribute sdl:+gl-doublebuffer+ 1)

  (let ((surface (sdl:set-video-mode width height bpp video-flags)))

    (when show-info

      (format t "OpenGL Information~%------------------~%Vendor:     ~A~%Version:    ~A~%Renderer:   ~A~%Extensions: ~A~%"
	      (gl:get-string gl:+vendor+)
	      (gl:get-string gl:+renderer+)
	      (gl:get-string gl:+version+)
	      (gl:get-string gl:+extensions+))
      (format t "~%")

      (format t "GL Rendering Context Bit Depths Obtained:~%")
      (multiple-value-bind (s r) (sdl:gl-get-attribute sdl:+gl-red-size+)
	(format t "- RED:   ~A ~A~%" s r))
      (multiple-value-bind (s g) (sdl:gl-get-attribute sdl:+gl-green-size+)
	(format t "- GREEN: ~A ~A~%" s g))
      (multiple-value-bind (s b) (sdl:gl-get-attribute sdl:+gl-blue-size+)
	(format t "- BLUE:  ~A ~A~%" s b))
      (multiple-value-bind (s d) (sdl:gl-get-attribute sdl:+gl-doublebuffer+)
	(format t "- DOUBLEBUFFER: ~A ~A~%" s (if d 'yes 'no)))
      (force-output))

    (sdl:wm-set-caption "LED cube modeller" nil)
    (gl:viewport 0 0 width height)
    ;; (gl:frustum (/ -640.0d0 480.0d0) (/ 640.0d0 480.0d0)
    ;;             1.0d0 -1.0d0 0.0d0 20.0d0)
    surface))
